import { keywords, schema } from "@arktype/schema";
import { printable, throwParseError } from "@arktype/util";
import { parseEntry } from "./shared.js";
const stringAndSymbolicEntriesOf = (o) => [
    ...Object.entries(o),
    ...Object.getOwnPropertySymbols(o).map((k) => [k, o[k]])
];
export const parseObjectLiteral = (def, ctx) => {
    const required = [];
    const optional = [];
    // We only allow a spread operator to be used as the first key in an object
    // because to match JS behavior any keys before the spread are overwritten
    // by the values in the target object, so there'd be no useful purpose in having it
    // anywhere except for the beginning.
    // Discussion in ArkType Discord:
    // https://discord.com/channels/957797212103016458/1103023445035462678/1182814502471860334
    const parsedEntries = stringAndSymbolicEntriesOf(def).map(parseEntry);
    if (parsedEntries[0].kind === "spread") {
        // remove the spread entry so we can iterate over the remaining entries
        // expecting non-spread entries
        const spreadEntry = parsedEntries.shift();
        const spreadNode = ctx.scope.parse(spreadEntry.innerValue, ctx);
        if (spreadNode.kind !== "intersection" ||
            !spreadNode.extends(keywords.object)) {
            return throwParseError(writeInvalidSpreadTypeMessage(printable(spreadEntry.innerValue)));
        }
        // TODO: move to props group merge in schema
        // For each key on spreadNode, add it to our object.
        // We filter out keys from the spreadNode that will be defined later on this same object
        // because the currently parsed definition will overwrite them.
        spreadNode.required?.forEach((spreadRequired) => !parsedEntries.some(({ innerKey }) => innerKey === spreadRequired.key) && required.push(spreadRequired));
        spreadNode.optional?.forEach((spreadOptional) => !parsedEntries.some(({ innerKey }) => innerKey === spreadOptional.key) && optional.push(spreadOptional));
    }
    for (const entry of parsedEntries) {
        if (entry.kind === "spread") {
            return throwParseError("Spread operator may only be used as the first key in an object");
        }
        ctx.path.push(`${typeof entry.innerKey === "symbol"
            ? `[${printable(entry.innerKey)}]`
            : entry.innerKey}`);
        const value = ctx.scope.parse(entry.innerValue, ctx);
        const inner = {
            key: entry.innerKey,
            value
        };
        if (entry.kind === "optional") {
            optional.push(inner);
        }
        else {
            required.push(inner);
        }
        ctx.path.pop();
    }
    return schema({
        basis: "object",
        required,
        optional
    });
};
export const writeInvalidPropertyKeyMessage = (indexDef) => `Indexed key definition '${indexDef}' must be a string, number or symbol`;
export const writeInvalidSpreadTypeMessage = (def) => `Spread operand must resolve to an object literal type (was ${def})`;
//# sourceMappingURL=objectLiteral.js.map